use regex::bytes::RegexBuilder;
use serde::Deserialize;
use thiserror::Error;

#[derive(Debug, Error)]
pub enum DeframeError {
    #[error("Start sequence not found")]
    StartSeqInvalid,
    #[error("Stop sequence not found")]
    StopSeqInvalid,
    #[error("Regex Error")]
    RegexError(#[from] regex::Error),
}

#[derive(Debug, Deserialize, Clone)]
#[serde(tag = "operation")]
pub enum DeframeOperation {
    StartSeq { start_seq: Vec<u8> },
    StopSeq { stop_seq: Vec<u8> },
    Replace { find: String, replace: Vec<u8> },
}

impl DeframeOperation {
    pub fn do_operation(&self, mut data: Vec<u8>) -> Result<Vec<u8>, DeframeError> {
        Ok(match self {
            DeframeOperation::StartSeq { start_seq } => {
                if data.starts_with(start_seq) {
                    data.drain(0..start_seq.len());

                    data
                } else {
                    return Err(DeframeError::StartSeqInvalid);
                }
            }
            DeframeOperation::StopSeq { stop_seq } => {
                if data.ends_with(stop_seq) {
                    data.drain(data.len() - stop_seq.len()..data.len());

                    data
                } else {
                    return Err(DeframeError::StopSeqInvalid);
                }
            }
            DeframeOperation::Replace { find, replace } => {
                let reg = RegexBuilder::new(find).unicode(false).build()?;

                reg.replace_all(&data, replace).to_vec()
            }
        })
    }
}

#[derive(Debug, Deserialize, Clone)]
pub struct Deframer {
    pub name: String,
    pub operations: Vec<DeframeOperation>,
}

impl Deframer {
    pub fn deframe(&self, mut data: Vec<u8>) -> Vec<u8> {
        for operation in &self.operations {
            data = operation.do_operation(data).unwrap();
        }

        data
    }
}

#[cfg(test)]
mod test {
    use crate::deframer::{DeframeOperation, Deframer};
    use crate::formatter::format_config::FormatConfig;

    #[test]
    fn test_start_seq() {
        let op = DeframeOperation::StartSeq {
            start_seq: vec![0x11, 0x00],
        };

        let data = vec![0x11, 0x00, 0x55, 0x55];
        let output = op.do_operation(data).unwrap();

        assert_eq!(output, vec![0x55, 0x55]);
    }

    #[test]
    fn test_stop_seq() {
        let op = DeframeOperation::StopSeq {
            stop_seq: vec![0x11, 0x00],
        };

        let data = vec![0x55, 0x55, 0x11, 0x00];
        let output = op.do_operation(data).unwrap();

        assert_eq!(output, vec![0x55, 0x55]);
    }

    #[test]
    fn test_replace_seq() {
        let op = DeframeOperation::Replace {
            find: "\x55\x11".to_string(),
            replace: vec![0x11],
        };

        let data = vec![0x55, 0x11, 0x55, 0x11, 0x00, 0x55];
        let output = op.do_operation(data).unwrap();

        assert_eq!(output, vec![0x11, 0x11, 0x00, 0x55]);
    }

    #[test]
    fn test_deframer() {
        let deframer = Deframer {
            name: "Test SLIP Deframmer".to_string(),
            operations: vec![
                DeframeOperation::StartSeq {
                    start_seq: vec![0xc0],
                },
                DeframeOperation::StopSeq {
                    stop_seq: vec![0xc0],
                },
                DeframeOperation::Replace {
                    find: r"\xDB\xDC".to_string(),
                    replace: vec![0xC0],
                },
                DeframeOperation::Replace {
                    find: r"\xDB\xDD".to_string(),
                    replace: vec![0xDB],
                },
            ],
        };

        let data = vec![0xC0, 0xDB, 0xDC, 0xDB, 0xDD, 0xC0];
        let output = deframer.deframe(data);

        assert_eq!(output, vec![0xc0, 0xDB]);
    }

    #[test]
    fn deserialize_deframer() {
        let config = FormatConfig::new(&None, &None).unwrap();
        let slip = config.get_deframer("slip").unwrap();

        assert_eq!(slip.name, "slip");

        assert_eq!(slip.operations.len(), 4);
    }
}
