use byteorder::{BigEndian, ByteOrder, LittleEndian};
use num_bigint::{BigInt, BigUint};

#[allow(dead_code)]
pub trait ByteOrderOperations: ByteOrder {
    fn last_byte(buf: &mut Vec<u8>) -> Option<&mut u8>;

    fn big_int(buf: &[u8], trim: usize) -> BigInt;

    fn big_uint(buf: &[u8], trim: usize) -> BigUint;

    fn big_u_int_to_bytes(big_int: BigUint) -> Vec<u8>;

    fn big_int_to_bytes(big_int: BigInt) -> Vec<u8>;

    fn pad_bytes(buf: &mut Vec<u8>, size: usize);

    fn flip() -> bool;
}

impl ByteOrderOperations for BigEndian {
    fn last_byte(buf: &mut Vec<u8>) -> Option<&mut u8> {
        buf.first_mut()
    }

    fn big_int(buf: &[u8], trim: usize) -> BigInt {
        let mut buf = buf.to_vec();
        let byte = buf.first_mut().unwrap();
        *byte = (*byte << trim) >> trim;
        BigInt::from_signed_bytes_be(&buf)
    }

    fn big_uint(buf: &[u8], trim: usize) -> BigUint {
        let mut buf = buf.to_vec();
        let byte = buf.first_mut().unwrap();
        *byte = (*byte << trim) >> trim;
        BigUint::from_bytes_be(&buf)
    }

    fn big_u_int_to_bytes(big_int: BigUint) -> Vec<u8> {
        big_int.to_bytes_be()
    }

    fn big_int_to_bytes(big_int: BigInt) -> Vec<u8> {
        big_int.to_signed_bytes_be()
    }

    fn pad_bytes(buf: &mut Vec<u8>, size: usize) {
        if size <= buf.len() {
            return;
        }

        let pad_to = size - buf.len();
        buf.splice(0..0, vec![0_u8; pad_to].iter().cloned());
    }

    fn flip() -> bool {
        true
    }
}

impl ByteOrderOperations for LittleEndian {
    fn last_byte(buf: &mut Vec<u8>) -> Option<&mut u8> {
        buf.last_mut()
    }

    fn big_int(buf: &[u8], trim: usize) -> BigInt {
        BigInt::from_signed_bytes_le(buf) >> trim
    }

    fn big_uint(buf: &[u8], trim: usize) -> BigUint {
        BigUint::from_bytes_le(buf) >> trim
    }

    fn big_u_int_to_bytes(big_int: BigUint) -> Vec<u8> {
        big_int.to_bytes_le()
    }

    fn big_int_to_bytes(big_int: BigInt) -> Vec<u8> {
        big_int.to_signed_bytes_le()
    }

    fn pad_bytes(buf: &mut Vec<u8>, size: usize) {
        if size <= buf.len() {
            return;
        }

        let pad_to = size - buf.len();

        let mut pad = vec![0_u8; pad_to];
        buf.append(&mut pad);
    }

    fn flip() -> bool {
        false
    }
}
