use crate::deframer::Deframer;
use crate::formatter::format::{Format, FormatError};
use crate::formatter::FormatConfigError;
use platform_dirs::AppDirs;
use rust_embed::RustEmbed;
use serde::Deserialize;
use std::fs::{read_dir, File};
use std::io::Read;
use std::path::PathBuf;
use std::{fs, str};

#[derive(RustEmbed)]
#[folder = "formats/"]
#[include = "*.toml"]
struct BuiltInFormats;

#[derive(Debug, Deserialize, Clone, Default)]
pub struct FormatConfig {
    #[serde(default)]
    pub deframers: Vec<Deframer>,
    #[serde(default)]
    pub formats: Vec<Format>,
}

impl FormatConfig {
    fn parse_config_file(
        src_file_name: &str,
        config_data: &str,
    ) -> Result<Self, FormatConfigError> {
        toml::from_str(config_data).map_err(|e| FormatConfigError::TomlError {
            file_name: src_file_name.to_string(),
            err: e,
        })
    }

    fn get_built_in_config(&mut self) -> Result<(), FormatConfigError> {
        for format_file_path in BuiltInFormats::iter() {
            if format_file_path.ends_with("md") {
                continue;
            }

            let format_file = BuiltInFormats::get(&format_file_path).unwrap();

            let config_str = str::from_utf8(&format_file.data).unwrap();

            let mut built_in: FormatConfig =
                Self::parse_config_file(&format_file_path, config_str)?;

            self.formats.append(&mut built_in.formats);
            self.deframers.append(&mut built_in.deframers)
        }

        Ok(())
    }

    fn get_file_config(&mut self, config_path: &Option<PathBuf>) -> Result<(), FormatConfigError> {
        if let Some(config_path) = config_path {
            let mut config_file = File::open(config_path)?;
            let mut config_data = String::new();
            config_file.read_to_string(&mut config_data)?;
            let mut arg_config =
                Self::parse_config_file(config_path.to_str().unwrap(), &config_data)?;

            self.formats.append(&mut arg_config.formats);
            self.deframers.append(&mut arg_config.deframers);
        }

        Ok(())
    }

    fn parse_directory(
        &mut self,
        directory_path: &Option<PathBuf>,
    ) -> Result<(), FormatConfigError> {
        if let Some(directory_path) = directory_path {
            for path in read_dir(directory_path)? {
                let file_name = path?.file_name();
                let config_file_path = directory_path.join(file_name);

                self.get_file_config(&Some(config_file_path))?;
            }
        }

        Ok(())
    }

    pub fn new(
        config_path: &Option<PathBuf>,
        global_config_path: &Option<PathBuf>,
    ) -> Result<Self, FormatConfigError> {
        let mut config = FormatConfig::default();

        config.get_built_in_config()?;

        config.get_file_config(config_path)?;

        let global_dir = match global_config_path {
            Some(g) => g.clone(),
            None => {
                let app_dirs = AppDirs::new(Some("formaty"), true).unwrap();

                if !app_dirs.config_dir.exists() {
                    fs::create_dir(&app_dirs.config_dir)?;
                }

                app_dirs.config_dir
            }
        };

        config.parse_directory(&Some(global_dir))?;

        Ok(config)
    }

    pub fn get_format(&self, name: &str) -> Result<Format, FormatError> {
        Ok(self
            .formats
            .iter()
            .find(|f| f.name.as_str() == name)
            .ok_or_else(|| FormatError::FormatNotFound(name.to_string()))?
            .clone())
    }

    pub fn get_deframer(&self, name: &str) -> Result<Deframer, FormatError> {
        Ok(self
            .deframers
            .iter()
            .find(|f| f.name.as_str() == name)
            .ok_or_else(|| FormatError::DeframmerNotFound(name.to_string()))?
            .clone())
    }
}
