pub mod bytes_operations;
pub mod format;
pub mod format_config;
mod printers;

use crate::formatter::format::FormatError;
use std::str;
use thiserror::Error;

#[derive(Debug, Error)]
#[allow(clippy::enum_variant_names)]
pub enum FormatConfigError {
    #[error("IO error")]
    IOError(#[from] std::io::Error),
    #[error("TOML Format Error")]
    TomlError {
        file_name: String,
        err: toml::de::Error,
    },
    #[error("UTF-8 String Decode Error")]
    Utf8Error(#[from] str::Utf8Error),
    #[error("Format Error")]
    FormatError(#[from] FormatError),
}

#[cfg(test)]
mod test {
    use crate::formatter::format::Format;
    use crate::formatter::format_config::FormatConfig;
    use crate::formatter::printers::print_bytes_as_array;

    #[derive(PartialOrd, PartialEq, Debug, Clone)]
    struct CCSDSPacket {
        packet_version: u8,
        packet_type: bool,
        sec_header_flag: bool,
        apid: u16,
        seq_flag: u8,
        seq: u16,
        packet_len: u16,
        data: Vec<u8>,
    }

    impl CCSDSPacket {
        fn from_bytes(data: &[u8]) -> Self {
            let packet_version = data[0] >> 5;
            let packet_type = ((data[0] >> 4) & 0x01) == 1;
            let sec_header_flag = ((data[0] >> 3) & 0x01) == 1;
            let apid = (((data[0] & 0x7) as u16) << 8) | (data[1] as u16);

            let seq_flag = data[2] >> 6;
            let seq = (((data[2] & 0x3F) as u16) << 8) | (data[3] as u16);

            let packet_len = ((data[4] as u16) << 8) | (data[5] as u16);

            CCSDSPacket {
                packet_version,
                packet_type,
                sec_header_flag,
                apid,
                seq_flag,
                seq,
                packet_len,
                data: data[6..].to_vec(),
            }
        }

        fn to_bytes(&self) -> Vec<u8> {
            let mut data = vec![0u8; 6];
            data[0] = (self.packet_version << 5)
                | ((self.packet_type as u8) << 4)
                | ((self.sec_header_flag as u8) << 3)
                | ((self.apid >> 8) as u8);
            data[1] = (self.apid & 0xFF) as u8;
            data[2] = (self.seq_flag << 6) | ((self.seq >> 8) as u8);
            data[3] = (self.seq & 0xFF) as u8;
            data[4] = (self.packet_len >> 8) as u8;
            data[5] = (self.packet_len & 0xff) as u8;

            data.append(&mut self.data.clone());
            data
        }

        // based off formats/ccsds.toml config
        fn print(&self) -> String {
            format!(
                "\
                Version Number: {}\n\
                Packet Type: {}\n\
                Secondary Header Flag: {}\n\
                APID: {:#x}\n\
                Sequence Flags: {}\n\
                Packet Sequence Count: {}\n\
                Data Length: {}\n\
                Data: {}\n",
                self.packet_version,
                self.packet_type as u8,
                self.sec_header_flag as u8,
                self.apid,
                self.seq_flag,
                self.seq,
                self.packet_len,
                print_bytes_as_array(&self.data)
            )
        }
    }

    #[test]
    fn test_ccsds() {
        let ccsds_packet = CCSDSPacket {
            packet_version: 0,
            packet_type: true,
            sec_header_flag: false,
            apid: 0x2FF,
            seq_flag: 3,
            seq: 0,
            packet_len: 5,
            data: vec![0x01, 0x02, 0x03, 0x04, 0x05],
        };

        let bytes = ccsds_packet.to_bytes();

        let parsed_packet = CCSDSPacket::from_bytes(&bytes);

        assert_eq!(ccsds_packet, parsed_packet);

        assert_eq!(bytes, parsed_packet.to_bytes())
    }

    fn get_ccsds_format() -> Format {
        let format_config = FormatConfig::new(&None, &None).unwrap();

        format_config.get_format("ccsds").unwrap()
    }

    #[test]
    fn test_ccsds_parse() {
        let ccsds_packet = CCSDSPacket {
            packet_version: 0,
            packet_type: true,
            sec_header_flag: false,
            apid: 0x2FF,
            seq_flag: 3,
            seq: 0,
            packet_len: 5,
            data: vec![0x01, 0x02, 0x03, 0x04, 0x05],
        };

        let format = get_ccsds_format();
        let data = ccsds_packet.to_bytes();
        let output = format
            .format_data(&data, &FormatConfig::default(), None)
            .unwrap();

        assert_eq!(ccsds_packet.print(), output)
    }

    #[test]
    fn test_ccsds_parse_hammer() {
        let format = get_ccsds_format();

        for apid in 0..(2u16.pow(11)) {
            let ccsds_packet = CCSDSPacket {
                packet_version: 0,
                packet_type: (apid % 2) == 0,
                sec_header_flag: (apid % 2) == 1,
                apid,
                seq_flag: (apid % 4) as u8,
                seq: (apid % 11),
                packet_len: 5,
                data: vec![0x01, 0x02, 0x03, 0x04, 0x05],
            };

            let data = ccsds_packet.to_bytes();
            let output = format
                .format_data(&data, &FormatConfig::default(), None)
                .unwrap();

            assert_eq!(ccsds_packet.print(), output)
        }
    }
}
